from pyearth._knot_search import (MultipleOutcomeDependentData,
                                  KnotSearchWorkingData,
                                  PredictorDependentData,
                                  KnotSearchReadOnlyData,
                                  KnotSearchData,
                                  knot_search,
                                  SingleWeightDependentData,
                                  SingleOutcomeDependentData)
from nose.tools import assert_equal
import numpy as np
from numpy.testing.utils import assert_almost_equal, assert_array_equal
from scipy.linalg import qr


def test_outcome_dependent_data():
    np.random.seed(10)
    m = 1000
    max_terms = 100
    y = np.random.normal(size=m)
    w = np.random.normal(size=m) ** 2
    weight = SingleWeightDependentData.alloc(w, m, max_terms, 1e-16)
    data = SingleOutcomeDependentData.alloc(y, weight, m, max_terms)

    # Test updating
    B = np.empty(shape=(m, max_terms))
    for k in range(max_terms):
        b = np.random.normal(size=m)
        B[:, k] = b
        code = weight.update_from_array(b)
        if k >= 99:
            1 + 1
        data.update()
        assert_equal(code, 0)
        assert_almost_equal(
            np.dot(weight.Q_t[:k + 1, :], np.transpose(weight.Q_t[:k + 1, :])),
            np.eye(k + 1))
    assert_equal(weight.update_from_array(b), -1)
#     data.update(1e-16)

    # Test downdating
    q = np.array(weight.Q_t).copy()
    theta = np.array(data.theta[:max_terms]).copy()
    weight.downdate()
    data.downdate()
    weight.update_from_array(b)
    data.update()
    assert_almost_equal(q, np.array(weight.Q_t))
    assert_almost_equal(theta, np.array(data.theta[:max_terms]))
    assert_almost_equal(
        np.array(data.theta[:max_terms]), np.dot(weight.Q_t, w * y))
    wB = B * w[:, None]
    Q, _ = qr(wB, pivoting=False, mode='economic')
    assert_almost_equal(np.abs(np.dot(weight.Q_t, Q)), np.eye(max_terms))

    # Test that reweighting works
    assert_equal(data.k, max_terms)
    w2 = np.random.normal(size=m) ** 2
    weight.reweight(w2, B, max_terms)
    data.synchronize()
    assert_equal(data.k, max_terms)
    w2B = B * w2[:, None]
    Q2, _ = qr(w2B, pivoting=False, mode='economic')
    assert_almost_equal(np.abs(np.dot(weight.Q_t, Q2)), np.eye(max_terms))
    assert_almost_equal(
        np.array(data.theta[:max_terms]), np.dot(weight.Q_t, w2 * y))


def test_knot_candidates():
    np.random.seed(10)
    m = 1000
    x = np.random.normal(size=m)
    p = np.random.normal(size=m)
    p[np.random.binomial(p=.1, n=1, size=m) == 1] = 0.
    x[np.random.binomial(p=.1, n=1, size=m) == 1] = 0.
    predictor = PredictorDependentData.alloc(x)
    candidates, candidates_idx = predictor.knot_candidates(
        p, 5, 10, 0, 0, set())
    assert_array_equal(candidates, x[candidates_idx])
    assert_equal(len(candidates), len(set(candidates)))
#     print candidates, np.sum(x==0)
#     print candidates_idx


def slow_knot_search(p, x, B, candidates, outcomes):
    # Brute force, utterly un-optimized knot search with no fast update.
    # Use only for testing the actual knot search function.
    # This version allows #for multiple outcome  columns.
    best_e = float('inf')
    best_k = 0
    best_knot = float('inf')
    for k, knot in enumerate(candidates):
        # Formulate the linear system for this candidate
        X = np.concatenate(
            [B, (p * np.maximum(x - knot, 0.0))[:, None]], axis=1)

        # Solve the system for each y and w
        e_squared = 0.0
        for y, w in outcomes:
            # Solve the system
            beta = np.linalg.lstsq(w[:, None] * X, w * y)[0]

            # Compute the error
            r = w * (y - np.dot(X, beta))
            e_squared += np.dot(r, r)
        # Compute loss
        e = e_squared  # / np.sum(w ** 2)

        # Compare to the best error
        if e < best_e:
            best_e = e
            best_k = k
            best_knot = knot
    return best_knot, best_k, best_e


def generate_problem(m, q, r, n_outcomes, shared_weight):
    # Generate some problem data
    x = np.random.normal(size=m)
    B = np.random.normal(size=(m, q))
    p = B[:, 1]
    knot = x[int(m / 2)]
    candidates = np.array(sorted(
        [knot] +
        list(x[np.random.randint(low=0, high=m, size=r - 1)])))[::-1]

    # These data need to be generated for each outcome
    outcomes = []
    if shared_weight:
        w = np.random.normal(size=m) ** 2
#         w = w * 0. + 1.
    for _ in range(n_outcomes):
        beta = np.random.normal(size=q + 1)
        y = (np.dot(
                np.concatenate([B, (p * np.maximum(x - knot, 0.0))[:, None]],
                               axis=1),
                beta) + 0.01 * np.random.normal(size=m))
        if not shared_weight:
            w = np.random.normal(size=m) ** 2
#             w = w * 0. + 1.
        outcomes.append((y, w))

    return x, B, p, knot, candidates, outcomes


def form_inputs(x, B, p, knot, candidates, y, w):
    # Formulate the inputs for the fast version
    m, q = B.shape
    max_terms = q + 2
    workings = []
    n_outcomes = w.shape[1]
    for _ in range(n_outcomes):
        working = KnotSearchWorkingData.alloc(max_terms)
        workings.append(working)
    outcome = MultipleOutcomeDependentData.alloc(
        y, w, m, n_outcomes, max_terms, 1e-16)
    for j in range(B.shape[1]):
        outcome.update_from_array(B[:, j])
    predictor = PredictorDependentData.alloc(x)
    constant = KnotSearchReadOnlyData(predictor, outcome)
    return KnotSearchData(constant, workings, q)


def test_knot_search():
    seed = 10
    np.random.seed(seed)
    m = 100
    q = 5
    r = 10
    n_outcomes = 3

    # Generate some problem data
    x, B, p, knot, candidates, outcomes = generate_problem(
        m, q,  r, n_outcomes, False)
    y = np.concatenate([y_[:, None] for y_, _ in outcomes], axis=1)
    w = np.concatenate([w_[:, None] for _, w_ in outcomes], axis=1)

    # Formulate the inputs for the fast version
    data = form_inputs(x, B, p, knot, candidates, y, w)

    # Get the answer using the slow version
    best_knot, best_k, best_e = slow_knot_search(p, x, B, candidates, outcomes)

    # Test the test
    assert_almost_equal(best_knot, knot)
    assert_equal(r, len(candidates))
    assert_equal(m, B.shape[0])
    assert_equal(q, B.shape[1])
    assert_equal(len(outcomes), n_outcomes)

    # Run fast knot search and compare results to slow knot search
    fast_best_knot, fast_best_k, fast_best_e = knot_search(data, candidates,
                                                           p, q, m, r,
                                                           len(outcomes), 0)
    assert_almost_equal(fast_best_knot, best_knot)
    assert_equal(candidates[fast_best_k], candidates[best_k])
    assert_almost_equal(fast_best_e, best_e)



from pyearth._basis import (Basis, ConstantBasisFunction, HingeBasisFunction,
                            LinearBasisFunction)
from pyearth.export import export_python_function, export_python_string,\
    export_sympy
from nose.tools import assert_almost_equal
import numpy
import six
from pyearth import Earth
from pyearth._types import BOOL
from pyearth.test.testing_utils import if_pandas,\
    if_sympy
from itertools import product
from numpy.testing.utils import assert_array_almost_equal

numpy.random.seed(0)

basis = Basis(10)
constant = ConstantBasisFunction()
basis.append(constant)
bf1 = HingeBasisFunction(constant, 0.1, 10, 1, False, 'x1')
bf2 = HingeBasisFunction(constant, 0.1, 10, 1, True, 'x1')
bf3 = LinearBasisFunction(bf1, 2, 'x2')
basis.append(bf1)
basis.append(bf2)
basis.append(bf3)
X = numpy.random.normal(size=(100, 10))
missing = numpy.zeros_like(X, dtype=BOOL)
B = numpy.empty(shape=(100, 4), dtype=numpy.float64)
basis.transform(X, missing, B)
beta = numpy.random.normal(size=4)
y = numpy.empty(shape=100, dtype=numpy.float64)
y[:] = numpy.dot(B, beta) + numpy.random.normal(size=100)
beta2 = numpy.random.normal(size=4)
y2 = numpy.empty(shape=100, dtype=numpy.float64)
y2[:] = numpy.dot(B, beta2) + numpy.random.normal(size=100)
Y = numpy.concatenate([y[:, None], y2[:, None]], axis=1)
default_params = {"penalty": 1}


def test_export_python_function():
    for smooth in (True, False):
        model = Earth(penalty=1, smooth=smooth, max_degree=2).fit(X, y)
        export_model = export_python_function(model)
        for exp_pred, model_pred in zip(model.predict(X), export_model(X)):
            assert_almost_equal(exp_pred, model_pred)


def test_export_python_string():
    for smooth in (True, False):
        model = Earth(penalty=1, smooth=smooth, max_degree=2).fit(X, y)
        export_model = export_python_string(model, 'my_test_model')
        six.exec_(export_model, globals())
        for exp_pred, model_pred in zip(model.predict(X), my_test_model(X)):
            assert_almost_equal(exp_pred, model_pred)

@if_pandas
@if_sympy
def test_export_sympy():
    import pandas as pd
    from sympy.utilities.lambdify import lambdify
    from sympy.printing.lambdarepr import NumPyPrinter

    class PyEarthNumpyPrinter(NumPyPrinter):
        def _print_Max(self, expr):
            return 'maximum(' + ','.join(self._print(i) for i in expr.args) + ')'

        def _print_NaNProtect(self, expr):
            return 'where(isnan(' + ','.join(self._print(a) for a in expr.args) + '), 0, ' \
                + ','.join(self._print(a) for a in expr.args) + ')'

        def _print_Missing(self, expr):
            return 'isnan(' + ','.join(self._print(a) for a in expr.args) + ').astype(float)'

    for smooth, n_cols, allow_missing in product((True, False), (1, 2), (True, False)):
        X_df = pd.DataFrame(X.copy(), columns=['x_%d' % i for i in range(X.shape[1])])
        y_df = pd.DataFrame(Y[:, :n_cols])
        if allow_missing:
            # Randomly remove some values so that the fitted model contains MissingnessBasisFunctions
            X_df['x_1'][numpy.random.binomial(n=1, p=.1, size=X_df.shape[0]).astype(bool)] = numpy.nan

        model = Earth(allow_missing=allow_missing, smooth=smooth, max_degree=2).fit(X_df, y_df)
        expressions = export_sympy(model) if n_cols > 1 else [export_sympy(model)]
        module_dict = {'select': numpy.select, 'less_equal': numpy.less_equal, 'isnan': numpy.isnan,
                       'greater_equal':numpy.greater_equal, 'logical_and': numpy.logical_and, 'less': numpy.less,
                       'logical_not':numpy.logical_not, "greater": numpy.greater, 'maximum':numpy.maximum,
                       'Missing': lambda x: numpy.isnan(x).astype(float),
                       'NaNProtect': lambda x: numpy.where(numpy.isnan(x), 0, x), 'nan': numpy.nan,
                       'float': float, 'where': numpy.where, 'numpy': numpy
                       }

        for i, expression in enumerate(expressions):
            # The lambdified functions for smoothed basis functions only work with modules='numpy' and
            # for regular basis functions with modules={'Max':numpy.maximum}.  This is a confusing situation
            func = lambdify(X_df.columns, expression, printer=PyEarthNumpyPrinter, modules=module_dict)
            y_pred_sympy = func(*[X_df.loc[:,var] for var in X_df.columns])

            y_pred = model.predict(X_df)[:,i] if n_cols > 1 else model.predict(X_df)
            assert_array_almost_equal(y_pred, y_pred_sympy)



cimport numpy as cnp
import numpy as np
from _types cimport FLOAT_t, INT_t, INDEX_t, BOOL_t
from _basis cimport Basis
from _record cimport ForwardPassRecord
from _knot_search cimport MultipleOutcomeDependentData

# cdef dict stopping_conditions

cdef class ForwardPasser:

    # User selected parameters
    cdef int endspan
    cdef int minspan
    cdef FLOAT_t endspan_alpha
    cdef FLOAT_t minspan_alpha
    cdef int max_terms
    cdef bint allow_linear
    cdef int max_degree
    cdef FLOAT_t thresh
    cdef FLOAT_t penalty
    cdef int check_every
    cdef int min_search_points
    cdef list xlabels
    cdef FLOAT_t zero_tol
    cdef list fast_heap
    cdef int use_fast
    cdef long fast_K
    cdef long fast_h
    cdef bint allow_missing
    cdef int verbose

    # Input data
    cdef cnp.ndarray X
    cdef cnp.ndarray missing
    cdef cnp.ndarray y
    cdef cnp.ndarray y_col_sum
    cdef cnp.ndarray y_row_sum
    cdef cnp.ndarray sample_weight
    cdef cnp.ndarray output_weight
    cdef INDEX_t m
    cdef INDEX_t n
    cdef FLOAT_t sst
    cdef FLOAT_t y_squared
    cdef FLOAT_t total_weight
    
    # Knot search data
    cdef MultipleOutcomeDependentData outcome
    cdef list predictors
    cdef list workings
    cdef INDEX_t n_outcomes
    
    # Working floating point data
    cdef cnp.ndarray B  # Data matrix in basis space
    cdef cnp.ndarray B_orth  # Orthogonalized version of B
    cdef cnp.ndarray c
    cdef cnp.ndarray c_sqr
    cdef cnp.ndarray norms
    cdef cnp.ndarray u
    cdef cnp.ndarray B_orth_times_parent_cum
    cdef FLOAT_t c_squared
    
    # Working integer data
    cdef cnp.ndarray sort_tracker
    cdef cnp.ndarray sorting
    cdef cnp.ndarray mwork
    cdef cnp.ndarray linear_variables
    cdef int iteration_number
    cdef cnp.ndarray has_missing
    
    # Object construction
    cdef ForwardPassRecord record
    cdef Basis basis

    cpdef Basis get_basis(ForwardPasser self)

    cpdef init_linear_variables(ForwardPasser self)

    cpdef run(ForwardPasser self)

    cdef stop_check(ForwardPasser self)

    cpdef orthonormal_update(ForwardPasser self, b)

    cpdef orthonormal_downdate(ForwardPasser self)

    cdef next_pair(ForwardPasser self)

#     cdef best_knot(ForwardPasser self, INDEX_t parent, cnp.ndarray[FLOAT_t, ndim=1] x,
#                    INDEX_t k, cnp.ndarray[INT_t, ndim=1] candidates,
#                    cnp.ndarray[INT_t, ndim=1] order,
#                    FLOAT_t * mse, FLOAT_t * knot,
#                    INDEX_t * knot_idx)



cimport numpy as cnp
from _types cimport FLOAT_t, INT_t, INDEX_t, BOOL_t
from _basis cimport Basis

cdef class Record:
    cdef list iterations
    cdef int num_samples
    cdef int num_variables
    cdef FLOAT_t penalty
    cdef FLOAT_t sst  # Sum of squares total

    cpdef append(Record self, Iteration iteration)

    cpdef FLOAT_t mse(Record self, INDEX_t iteration)

    cpdef FLOAT_t rsq(Record self, INDEX_t iteration)

    cpdef FLOAT_t gcv(Record self, INDEX_t iteration)

    cpdef FLOAT_t grsq(Record self, INDEX_t iteration)

cdef class PruningPassRecord(Record):
    cdef readonly INDEX_t selected

    cpdef set_selected(PruningPassRecord self, INDEX_t selected)

    cpdef INDEX_t get_selected(PruningPassRecord self)

    cpdef roll_back(PruningPassRecord self, Basis basis)

cdef class ForwardPassRecord(Record):
    cdef readonly int stopping_condition

    cdef list xlabels

    cpdef set_stopping_condition(ForwardPassRecord self, int stopping_condition)

cdef class Iteration:
    cdef FLOAT_t mse
    cdef INDEX_t size

    cpdef FLOAT_t get_mse(Iteration self)

    cpdef INDEX_t get_size(Iteration self)

cdef class PruningPassIteration(Iteration):
    cdef INDEX_t pruned

    cpdef INDEX_t get_pruned(PruningPassIteration self)

cdef class FirstPruningPassIteration(PruningPassIteration):
    pass

cdef class ForwardPassIteration(Iteration):
    cdef INDEX_t parent
    cdef INDEX_t variable
    cdef FLOAT_t knot
    cdef int code
    cdef bint no_candidates

    cpdef set_no_candidates(ForwardPassIteration self, bint value)

    cpdef no_further_candidates(ForwardPassIteration self)

cdef class FirstForwardPassIteration(ForwardPassIteration):
    cpdef INDEX_t get_size(FirstForwardPassIteration self)



import os
import numpy


class BaseContainer(object):
    filename = os.path.join(os.path.dirname(__file__), '../test_data.csv')
    data = numpy.genfromtxt(filename, delimiter=',', skip_header=1)
    X = numpy.array(data[:, 0:5])
    y = numpy.array(data[:, 5])



from cython cimport view
from _types cimport FLOAT_t, INT_t, INDEX_t, BOOL_t

cdef class UpdatingQT:
    cdef readonly int m
    cdef readonly int max_n
    cdef readonly Householder householder
    cdef readonly int k
    cdef readonly FLOAT_t[::1, :] Q_t
    cdef readonly FLOAT_t zero_tol
    cdef readonly BOOL_t[::1] dependent_cols
    cpdef void update_qt(UpdatingQT self, bint dependent)
    cpdef void update(UpdatingQT self, FLOAT_t[:] x)
    cpdef void downdate(UpdatingQT self)
    cpdef void reset(UpdatingQT self)

cdef class Householder:
    cdef readonly int k
    cdef readonly int m
    cdef readonly int max_n
    cdef readonly FLOAT_t[::1, :] V
    cdef readonly FLOAT_t[::1, :] T
    cdef readonly FLOAT_t[::1] tau
    cdef readonly FLOAT_t[::1] beta
    cdef readonly FLOAT_t[::1, :] work
    cdef readonly FLOAT_t zero_tol
    cpdef void downdate(Householder self)
    cpdef void reset(Householder self)
    cpdef bint update_from_column(Householder self, FLOAT_t[:] c)
    cpdef bint update_v_t(Householder self)
    cpdef void left_apply(Householder self, FLOAT_t[::1, :] C)
    cpdef void left_apply_transpose(Householder self, FLOAT_t[::1, :] C)
    cpdef void right_apply(Householder self, FLOAT_t[::1, :] C)
    cpdef void right_apply_transpose(Householder self, FLOAT_t[::1, :] C)



'''
Created on Feb 16, 2013

@author: jasonrudy
'''

import os
import numpy

from nose.tools import assert_equal

from pyearth._forward import ForwardPasser
from pyearth._basis import (Basis, ConstantBasisFunction,
                            HingeBasisFunction, LinearBasisFunction)
from pyearth._types import BOOL

numpy.random.seed(0)
basis = Basis(10)
constant = ConstantBasisFunction()
basis.append(constant)
bf1 = HingeBasisFunction(constant, 0.1, 10, 1, False, 'x1')
bf2 = HingeBasisFunction(constant, 0.1, 10, 1, True, 'x1')
bf3 = LinearBasisFunction(bf1, 2, 'x2')
basis.append(bf1)
basis.append(bf2)
basis.append(bf3)
X = numpy.random.normal(size=(100, 10))
missing = numpy.zeros_like(X).astype(BOOL)
B = numpy.empty(shape=(100, 4), dtype=numpy.float64)
basis.transform(X, missing, B)
beta = numpy.random.normal(size=4)
y = numpy.empty(shape=100, dtype=numpy.float64)
y[:] = numpy.dot(B, beta) + numpy.random.normal(size=100)
sample_weight = numpy.ones((X.shape[0], 1))


def test_run():

    forwardPasser = ForwardPasser(X, missing, y[:, numpy.newaxis],
                                  sample_weight,
                                  max_terms=1000, penalty=1)

    forwardPasser.run()
    res = str(forwardPasser.get_basis()) + \
        '\n' + str(forwardPasser.trace())
    filename = os.path.join(os.path.dirname(__file__),
                            'forward_regress.txt')
#     with open(filename, 'w') as fl:
#         fl.write(res)
    with open(filename, 'r') as fl:
        prev = fl.read()
    assert_equal(res, prev)
